#include "mainwindow.h"
#include "./ui_mainwindow.h"
#include <memory>
#include <fstream>
#include <chrono>
#include <thread>

#include "mapviewer.h"
#include "include/slam_manager.h"
#include "Inspvaxa.h"
#include "can2swc.h"
#include "functions.h"
#include "include/data/data_coordinates.h"

MapViewer map_;
std::shared_ptr<slam::SlamManager> p_slam;
InspvaxaData ins_converter;
std::string file_path;
bool stoped{false};
std::vector<CAN2SWC> can_data_stream;

vector<string> split(string str, char delimiter) {
    vector<string> tokens;
    string token;
    istringstream ss(str);

    while (getline(ss, token, delimiter)) {
        tokens.push_back(token);
    }

    return tokens;
}

void package_ins_inputs(slam::InsData *const _insData, CAN2SWC const _canData)
{
    _insData->utc_timestamp = _canData.ins_sol_result_longitude_signals.timestamp_s * 1e3 +
            _canData.ins_sol_result_longitude_signals.timestamp_ns * 1e-6; // milliseconds

    _insData->longitude = _canData.ins_sol_result_longitude_signals.ins_longitude;
    _insData->latitude = _canData.ins_sol_result_latitude_signals.ins_latitude;
    _insData->heading = _canData.ins_sol_result_angle_signals.ins_yaw_angle;

    std::cout << "input gps:" << _insData->longitude << "," << _insData->latitude << std::endl;
    //_insData->status
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    map_.Init(ui->customPlot, ui->customPlot_2);

    //    p_slam = std::make_shared<slam::SlamManager>();
    //    p_slam->Init();

    QString basePath = QCoreApplication::applicationDirPath();
    basePath.replace(basePath.length() - 6, 6, "");


    qRegisterMetaType<slam::slam_road_model >("slam::slam_road_model ");//注册该类型
    QObject::connect(this, &MainWindow::send_rm_outputs,
                     this, &MainWindow::plot_rm_outputs);

    QObject::connect(this, &MainWindow::send_odo,
                     this, &MainWindow::plot_odo);

    qRegisterMetaType<std::vector<slam::Vector2d>>("std::vector<slam::Vector2d>");
    qRegisterMetaType<std::vector<slam::Vector3d>>("std::vector<slam::Vector3d>");
    QObject::connect(this, &MainWindow::send_ref,
                     this, &MainWindow::plot_ref);
    QObject::connect(this, &MainWindow::send_ref_c,
                     this, &MainWindow::plot_ref_c);


    file_path = basePath.toStdString()+ "data/ins_can_210331.csv";
    //

    //file_path = basePath.toStdString()+ "data/CAN2SWCEvent.csv";

    start();

    //parseCSV();
    //ReadInsCAN();

    //ReplayTrajectory();
}

void MainWindow::parseCSV() {
    std::ifstream file(file_path);
    std::string line;

    if (!file.is_open()) {
        std::cerr << "Failed to open file" << std::endl;
        return;
    }

    // Read the header line first if needed
    std::getline(file, line);

    uint32_t slam_run_counter = 0;

    slam::PackageLine wk_package_line;

    slam::slam_road_model wk_rm_;
    slam::slam_odometry wk_odom;
    slam::ControlWaypoints wk_cont_way_points;

//    wk_rm_.lanes.resize(1);
//    wk_rm_.lines.resize(3);
//    wk_rm_.lines[0].points.resize(200);

    // Read each line from the file
    while (getline(file, line)) {
        std::stringstream linestream(line);
        std::string cell;

        CAN2SWC data;

        // Read each value from the line
        std::getline(linestream, cell, ','); // Read time
        data.time = std::stoll(cell);

        // Skip unneeded fields (assuming they are between 'time' and the fields we care about)
        for (int i = 0; i < 63; ++i) {
            std::getline(linestream, cell, ','); // Just read and ignore these fields
        }

        // INS_SOL_RESULT_ANGLE_Signals
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.ins_course_angle = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.ins_yaw_angle = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.ins_pitch = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.ins_roll = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.timestamp_s = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.timestamp_ns = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_angle_signals.crc8 = std::stoi(cell);

        // INS_SOL_RESULT_ALTITUDE_Signals
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.ins_altitude = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.ins_altitude_std = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.ins_undulation = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.timestamp_s = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.timestamp_ns = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_altitude_signals.crc8 = std::stoi(cell);

        // INS_SOL_RESULT_LONGITUDE_Signals
        std::getline(linestream, cell, ','); data.ins_sol_result_longitude_signals.ins_longitude = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_longitude_signals.ins_longitude_std = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_longitude_signals.timestamp_s = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_longitude_signals.timestamp_ns = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_longitude_signals.crc8 = std::stoi(cell);

        // INS_SOL_RESULT_LATITUDE_Signals
        std::getline(linestream, cell, ','); data.ins_sol_result_latitude_signals.ins_latitude = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_latitude_signals.ins_latitude_std = std::stod(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_latitude_signals.timestamp_s = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_latitude_signals.timestamp_ns = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_latitude_signals.crc8 = std::stoi(cell);

        // INS_SOL_RESULT_COMMON_Signals
        std::getline(linestream, cell, ','); data.ins_sol_result_common_signals.ins_gps_week = std::stoi(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_common_signals.ins_gps_sow = std::stoi(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_common_signals.timestamp_s = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_common_signals.timestamp_ns = std::stol(cell);
        std::getline(linestream, cell, ','); data.ins_sol_result_common_signals.crc8 = std::stoi(cell);


        // Add the filled structure to the vector
        can_data.push_back(data);

        slam::InsData _insdata;

        package_ins_inputs(&_insdata, data);

        // 重置wk_rm_的状态
        wk_rm_.lanes.clear();   // 清除车道信息
        wk_rm_.lines.clear();   // 清除线条信息
        wk_rm_.lanes.resize(1); // 重新设定车道数量
        wk_rm_.lines.resize(3); // 重新设定线条数量

        // 初始化线条中的点
        for (int i = 0; i < 3; ++i)
        {
            wk_rm_.lines[i].points.resize(100); // 每条线有100个点
        }

        if (slam_run_counter == 0)
        {
            wk_package_line.Init();
            printf("slam initialized!\n");
            std::vector<slam::Vector2d> ref_points = wk_package_line.GetRefLine();
            emit send_ref(ref_points);
        }

        wk_package_line.SetHostVehiclePosition(_insdata);
        bool wk_odom_flag = wk_package_line.PackageOdom(wk_odom);
        bool wk_rm_flag = wk_package_line.PackageRM(wk_rm_.lines[0].points);
        wk_rm_.lanes[0].lane_id = 1;
        wk_rm_.lanes[0].lane_no = 2;
        wk_rm_.lanes[0].center_line_id = 10;
        wk_rm_.lanes[0].left_line_id = 11;
        wk_rm_.lanes[0].right_line_id = 12;

        wk_rm_.lanes[0].speed_limit_value.max_speed_limit = 50;
        wk_rm_.lanes[0].speed_limit_value.min_speed_limit = 10;

        wk_rm_.lines[0].id = 10;
        wk_rm_.lines[1].id = 11;
        wk_rm_.lines[2].id = 12;

        wk_rm_.lines[1].points.resize(wk_rm_.lines[0].points.size());
        wk_rm_.lines[2].points.resize(wk_rm_.lines[0].points.size());

        for (size_t i = 0; i < wk_rm_.lines[0].points.size(); i++)
        {
            wk_rm_.lines[1].points[i].point.x = wk_rm_.lines[0].points[i].point.x - 1.8f;
            wk_rm_.lines[1].points[i].point.y = wk_rm_.lines[0].points[i].point.y ;

            wk_rm_.lines[2].points[i].point.x = wk_rm_.lines[0].points[i].point.x + 1.8f;
            wk_rm_.lines[2].points[i].point.y = wk_rm_.lines[0].points[i].point.y ;
        }

        emit send_ref_c(wk_package_line.c_line);

        slam_run_counter++;
        //std::cout << index++ << std::endl;
        emit send_rm_outputs(wk_rm_);
        emit send_odo(wk_odom.position.position.x, wk_odom.position.position.y, _insdata.heading);
        while(stoped)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }

    file.close();
}

void MainWindow::plot_ref_c(std::vector<slam::Vector3d> _line)
{
    QVector<QCPCurveData> dataSpiral;

    for(uint j = 0; j < _line.size(); ++j)
    {
        auto _p{_line[j]};

        dataSpiral.push_back(QCPCurveData(j,_p.x,_p.y));

    }

    map_.PlotLine_c(dataSpiral, QColor(255,0,0), 5);
}

void MainWindow::plot_ref(std::vector<slam::Vector2d> _line)
{
    QVector<QCPCurveData> dataSpiral;

    for(uint j = 0; j < _line.size(); ++j)
    {
        auto _p{_line[j]};

        dataSpiral.push_back(QCPCurveData(j,_p.x,_p.y));

    }

    map_.PlotLine(dataSpiral, QColor(255,255,0), 2);
    //map_.PlotPoints(dataSpiral, QColor(255,0,255));
}

void MainWindow::start()
{
    std::thread pack_thread(&MainWindow::ReadInsCAN, this);
    pack_thread.detach();

    //        std::thread pack_thread(&MainWindow::parseCSV, this);
    //        pack_thread.detach();
}

void MainWindow::ReplayTrajectory()
{

}


MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::plot_rm_outputs(slam::slam_road_model in_rmData)
{

    for (uint i = 0; i < in_rmData.lines.size(); ++i)
    {
        QVector<QCPCurveData> dataSpiral;

        for(uint j = 0; j < in_rmData.lines[i].points.size(); ++j)
        {
            auto _p{in_rmData.lines[i].points[j].point};

            dataSpiral.push_back(QCPCurveData(j,_p.x,_p.y));

        }

        map_.PlotLines(dataSpiral, QColor(0,255,0), 2, i);
    }

    double c0 = -5.3058;
    double c1 = 0.512311;
    double c2 =  -0.000704878;
    double c3 = 4.89008e-06;

    QVector<QCPCurveData> dataSpiral;
    double x0 = -10;
    for(uint j = 0; j < 100; ++j)
    {
        double x = x0 + j*0.5;
        double y = c0 + x*c1 + x*x*c2 + x*x*x*c3;
        dataSpiral.push_back(QCPCurveData(j,-y,x));
    }

    //map_.PlotLines(dataSpiral, QColor(255,0,0), 2, 1);
}

void MainWindow::plot_odo(double x, double y, double heading)
{
    map_.plotOdo(x,y, heading*M_PI/180);
}

void MainWindow::ReadInsCAN()
{
    std::string line;
    auto fp_{file_path};
    std::ifstream ins_file(fp_);

    if (!ins_file.is_open())
    {
        std::cout << "无法打开文件" << std::endl;
        return;
    }

    int index = 0;

    uint32_t slam_run_counter = 0;

    slam::PackageLine wk_package_line;

    slam::slam_road_model wk_rm_;
    slam::slam_odometry wk_odom;
    slam::ControlWaypoints wk_cont_way_points;

//    wk_rm_.lanes.resize(1);
//    wk_rm_.lines.resize(3);
//    wk_rm_.lines[0].points.resize(200);


    while (std::getline(ins_file, line))
    {
        vector<string> tokens = split(line, ',');
        slam::InsData _insdata;

        _insdata.latitude = std::stod(tokens[1]);
        _insdata.longitude = std::stod(tokens[2]);
        _insdata.heading = std::stod(tokens[3]);

        //        _insdata.latitude = 31.8346197;
        //        _insdata.longitude = 117.0949578;
        //        _insdata.heading = 150.3600006;

        _insdata.utc_timestamp = std::stod(tokens[0])*1000;

        // 重置wk_rm_的状态
        wk_rm_.lanes.clear();   // 清除车道信息
        wk_rm_.lines.clear();   // 清除线条信息
        wk_rm_.lanes.resize(1); // 重新设定车道数量
        wk_rm_.lines.resize(3); // 重新设定线条数量

        // 初始化线条中的点
        for (int i = 0; i < 3; ++i)
        {
            wk_rm_.lines[i].points.resize(100); // 每条线有100个点
        }

        if (slam_run_counter == 0)
        {
            wk_package_line.Init();
            printf("slam initialized!\n");
            std::vector<slam::Vector2d> ref_points = wk_package_line.GetRefLine();
            emit send_ref(ref_points);
        }

        wk_package_line.SetHostVehiclePosition(_insdata);
        bool wk_odom_flag = wk_package_line.PackageOdom(wk_odom);
        bool wk_rm_flag = wk_package_line.PackageRM(wk_rm_.lines[0].points);
        wk_rm_.lanes[0].lane_id = 1;
        wk_rm_.lanes[0].lane_no = 2;
        wk_rm_.lanes[0].center_line_id = 10;
        wk_rm_.lanes[0].left_line_id = 11;
        wk_rm_.lanes[0].right_line_id = 12;

        wk_rm_.lanes[0].speed_limit_value.max_speed_limit = 50;
        wk_rm_.lanes[0].speed_limit_value.min_speed_limit = 10;

        wk_rm_.lines[0].id = 10;
        wk_rm_.lines[1].id = 11;
        wk_rm_.lines[2].id = 12;

        wk_rm_.lines[1].points.resize(wk_rm_.lines[0].points.size());
        wk_rm_.lines[2].points.resize(wk_rm_.lines[0].points.size());

        for (size_t i = 0; i < wk_rm_.lines[0].points.size(); i++)
        {
            wk_rm_.lines[1].points[i].point.x = wk_rm_.lines[0].points[i].point.x - 1.8f;
            wk_rm_.lines[1].points[i].point.y = wk_rm_.lines[0].points[i].point.y ;

            wk_rm_.lines[2].points[i].point.x = wk_rm_.lines[0].points[i].point.x + 1.8f;
            wk_rm_.lines[2].points[i].point.y = wk_rm_.lines[0].points[i].point.y ;
        }

        emit send_ref_c(wk_package_line.c_line);

        slam_run_counter++;
        //std::cout << index++ << std::endl;
        emit send_rm_outputs(wk_rm_);
        emit send_odo(wk_odom.position.position.x, wk_odom.position.position.y, _insdata.heading);
        while(stoped)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
        std::this_thread::sleep_for(std::chrono::milliseconds(100));
    }

    //plot_rm_outputs(wk_rm_);
    ins_file.close();
}

void MainWindow::ReadInsData(std::string fp_)
{
    std::string line;
    std::ifstream ins_file(fp_);

    if (!ins_file.is_open())
    {
        std::cout << "无法打开文件" << std::endl;
        return;
    }

    int index = 0;
    while (std::getline(ins_file, line))
    {
        if (line.substr(0, 9) == "#INSPVAXA") {

            INSPVAXA _inspvaxa = ins_converter.parseINSPVAXA(line);
            slam::InsData _insdata = ins_converter.MapINSPVAXAToInsData(_inspvaxa);

            p_slam->InputInsData2Slam(_insdata);

            //std::this_thread::sleep_for(std::chrono::milliseconds(1));

            std::cout << index++ << std::endl;
        }
    }

    ins_file.close();
}

void MainWindow::on_actionplay_triggered()
{

}

void MainWindow::on_actionfocus_triggered()
{
    map_.SetOdomFocus();
}

void MainWindow::on_action3DQuat_triggered()
{

}

void MainWindow::on_actionplay_toggled(bool arg1)
{
    stoped = !arg1;
}
